'use strict';

//module imports
const FormeValidationError = require('../../../errors').FormeValidationError;
const validator = require('validator');

//local imports
const InputValidateHandler = require('../inputValidateHandler');

//main class
class InputValidateHandlerIs extends InputValidateHandler {
    constructor(type, options, error) {
        super(error);
        this.type = type.toLowerCase();
        this.options = options;
    }

    //properties
    get configuration() {
        const configuration = super.configuration;
        configuration.type = this.type;
        if (this.options) {
            configuration.options = this.options;
        }
        return configuration;
    }

    get htmlInputType() {
        switch(this.type) {
            case 'tel':
                return 'tel';
            case 'color':
                return 'color';
            case 'email':
                return 'email';
            case 'isodate':
                return 'date';
            case 'subdomain':
                return 'text';
            case 'float':
            case 'decimal':
            case 'number':
                return 'number';
            default:
                //unhandled types dont override
                return null;
        }
    }

    //api
    execute(input, state) {
        //we take into account the state.require flag
        if ((state.value !== null && state.value !== '') || state.require) {
            //lets convert our type into https://www.npmjs.com/package/validator#validators
            switch (this.type) {
                case 'uk-postcode':
                    if (!state.value || !/[A-Z]{1,2}[0-9]{1,2}[A-Z]? ?[0-9][A-Z]{2}/i.test(state.value)) {
                        return Promise.reject(new FormeValidationError('{label} is not a valid UK postcode'));
                    }
                    break;

                case 'alphanumeric':
                    //noinspection Annotator
                    if (!state.value || !validator.isAlphanumeric(state.value, this.options)) {
                        return Promise.reject('{label} is not alphanumeric');
                    }
                    break;

                case 'email':
                    //noinspection Annotator
                    if (!state.value || !validator.isEmail(state.value, this.options)) {
                        return Promise.reject(new FormeValidationError('{label} is not a valid email'));
                    }
                    break;

                case 'username':
                    //noinspection Annotator
                    if (!state.value || validator.whitelist(state.value, 'a-zA-Z0-9\_\-') !== state.value) {
                        return Promise.reject(new FormeValidationError('{label} is not a valid username'));
                    }
                    break;

                case 'subdomain':
                    //noinspection Annotator
                    if (!state.value || validator.whitelist(state.value, 'a-zA-Z0-9\_\-') !== state.value) {
                        return Promise.reject(new FormeValidationError('{label} is not a valid subdomain'));
                    }
                    break;

                case 'isodate':
                    //noinspection Annotator
                    if (!state.value || !validator.isISO8601(state.value)) {
                        return Promise.reject(new FormeValidationError('{label} is not a valid date'));
                    }
                    break;

                case 'text':
                    //noinspection Annotator
                    if (!state.value || !validator.isAscii(state.value)) {
                        return Promise.reject(new FormeValidationError('{label} does not contain valid text'));
                    }
                    break;

                case 'boolean':
                case 'bool':
                    //noinspection Annotator,Annotator,Annotator
                    if (state.value === null || state.value === undefined || (typeof state.value !== 'string' && typeof state.value !== 'boolean') || (typeof state.value === 'string' && !validator.isInt(state.value))) {
                        return Promise.reject(new FormeValidationError('{label} is not a valid boolean'));
                    }
                    break;

                case 'int':
                    //noinspection Annotator
                    if (state.value === null || state.value === undefined || (typeof state.value !== 'string' && typeof state.value !== 'number') || (typeof state.value === 'string' && !validator.isInt(state.value))) {
                        return Promise.reject(new FormeValidationError('{label} is not a valid integer'));
                    }
                    break;

                case 'float':
                    //noinspection Annotator
                    if (state.value === null || state.value === undefined || (typeof state.value !== 'string' && typeof state.value !== 'number') || (typeof state.value === 'string' && validator.isFloat(state.value))) {
                        return Promise.reject(new FormeValidationError('{label} is not a valid float'));
                    }
                    break;

                case 'decimal':
                case 'number':
                    //noinspection Annotator
                    if (state.value === null || state.value === undefined || (typeof state.value !== 'string' && typeof state.value !== 'number') || (typeof state.value === 'string' && !validator.isDecimal(state.value))) {
                        return Promise.reject(new FormeValidationError('{label} is not a valid number'));
                    }
                    break;

                case 'color':
                case 'colour':
                    //noinspection Annotator
                    if (!state.value || !validator.isHexColor(state.value)) {
                        return Promise.reject(new FormeValidationError('{label} is not a valid color'));
                    }
                    break;

                case 'phone':
                case 'telephone':
                case 'tel':
                    //noinspection Annotator
                    if (!state.value || validator.isWhitelisted(state.value, '0-9\(\)\-\s\+')) {
                        return Promise.reject(new FormeValidationError('{label} is not a valid phone number'));
                    }
                    break;

                case 'uuid'://defaults to UUID4
                case 'uuid4':
                    //noinspection Annotator
                    if (!state.value || validator.idUUID(state.value, 4)) {
                        return Promise.reject(new FormeValidationError('{label} is not a valid UUID'));
                    }
                    break;
                case 'uuid3':
                    //noinspection Annotator
                    if (!state.value || validator.idUUID(state.value, 3)) {
                        return Promise.reject(new FormeValidationError('{label} is not a valid UUID'));
                    }
                    break;
                case 'uuid5':
                    //noinspection Annotator
                    if (!state.value || validator.idUUID(state.value, 5)) {
                        return Promise.reject(new FormeValidationError('{label} is not a valid UUID'));
                    }
                    break;

                default:
                    //handle any others
                    if (this.type.length === 0) {
                        return Promise.reject(new FormeValidationError('{label} is not valid'));
                    } else {
                        const func = validator['is' + this.type[0].toUpperCase() + this.type.slice(1).toLowerCase()];
                        if (func === undefined) {
                            return Promise.reject(new FormeValidationError('{label} has an unknown type'));
                        } else {
                            //try calling validator's custom is routine
                            if (state.value === null || state.value === undefined || !func(state.value, this.options)) {
                                return Promise.reject(new FormeValidationError('{label} is not a valid '+this.type));
                            }
                        }
                    }
                    break;
            }
        }

        //skip
        return Promise.resolve();
    }
}

//expose module
module.exports = InputValidateHandlerIs;